/* INT10/raw screen handling for DOSRUN project
 * (C)1999 LGB of DC, Gabor Lenart, lgb@vlug.vein.hu
 * GNU/GPL code
 */

#include "config_dosrun.h"
#ifdef CONFIG_INT10   /* a BIG ifdef block */

#include <stdio.h>
#include "dosrun.h"
#include "screen.h"
#include "options.h"


extern word oldcs,oldip;
int curx=0,cury=0,curattr=7;

#define SCROLL_SCREEN scrolldown_window(0,0,scrxsize-1,scrysize-1,7,1)
#define CLEAR_SCREEN delete_window(0,0,scrxsize-1,scrysize-1,7)


static void delete_window ( int x1 , int y1 , int x2 , int y2 , int attr )
{
  int x,y;
	word *p=(word*)0xB8000+y1*scrxsize;
  for (y=y1;y<=y2;y++) {
    for (x=x1;x<=x2;x++)
      p[x]=' '+(attr<<8);
		p+=scrxsize;
  }
}


static void scrolldown_window ( int x1 , int y1 , int x2 , int y2 , int attr , int n )
{
  if (x1>=scrxsize||y1>=scrysize||x2>=scrxsize||y2>=scrysize||x1>x2||y2>y2) return;
  if (!n||n>y2-y1) delete_window(x1,y1,x2,y2,attr); else {
    int y;
    word *p=(word*)0xB8000+x1+y2*scrxsize;
    for (y=y1;y<=y2-n;y++) {
      memcpy(p,p-n*scrxsize,(x2-x1+1)<<1);
      p-=scrxsize;
    }
    delete_window(x1,y1,x2,y1+n-1,attr);
  }
}


static void scrollup_window ( int x1 , int y1 , int x2 , int y2 , int attr , int n )
{
  if (x1>=scrxsize||y1>=scrysize||x2>=scrxsize||y2>=scrysize||x1>x2||y1>y2) return;
  if (!n||n>y2-y1) delete_window(x1,y1,x2,y2,attr); else {
    int y;
    word *p=(word*)0xB8000+x1+y1*scrxsize;
    for (y=y1;y<=y2-n;y++) {
      memcpy(p,p+n*scrxsize,(x2-x1+1)<<1);
      p+=scrxsize;
    }
    delete_window(x1,y2-n+1,x2,y2,attr);
  }
}

/* These two functions below is the heart of char output inside dosrun. 
 * you MUST call this function to display character on terminal including
 * writing into dos stdout for example. It can work in both of full screen and
 * terminal mode. `col' is color (only has meaning in full screen mode)
 * and can be -1 to ignore this information. Cursor (full screen mode) goes
 * to the next position including screen scrolling if it's needed.
 * Control characters are supported, possible character and controll
 * code conversion (dos->unix) is done - or at least can be done - here, too
 * for TERMINAL mode !! For fullscreen mode, conversion is done by screen
 * updater thread in source file `screen.c'.
 */
static void screen_raw_char ( int c , int col )  /* for FULLSCREEN mode */
{
  switch (c) {
	  case 10 : curx=0;
				      return;
		case 13 : if (cury==scrysize-1) SCROLL_SCREEN; else cury++;
				      return;
		case  7 : putchar(7);
				      fflush(stdout);
				      return;
		default : if (col==-1) *((byte*)0xB8000+((curx+cury*scrxsize)<<1))=c; else *((word*)0xB8000+curx+cury*scrxsize)=c+(col<<8);
				      if (curx==scrxsize-1) {
					      curx=0;
					      if (cury==scrysize-1) SCROLL_SCREEN; else cury++;
				      } else curx++;
				      return;
  }
}

		
void screen_char ( int c , int col )  /* for both of fullscreen and terminal mode */
{
	if (config.fullscreen) screen_raw_char(c,col); else {
		putchar(c);
		fflush(stdout);
  }
}


/* emulation code for INT10, 0x13 : string wr */
static __inline__ void int10_13_emu ( void )
{
  byte *p=DOS_PB(ES,BP);
	word *w;
	int n=CX;
	if (DL>scrxsize||DH>scrysize||AL>3) return;
	if (AL&1) {
		curx=DL;
		cury=DH;
	} else w=(word*)0xB8000+DL+DH*scrxsize;
	while (n--) {
		if (AL&1) screen_char(*p,AL&2?p[1]:BL); else *(w++)=*p+((AL&2?p[1]:BL)<<8);
		p+=AL&2?2:1;
	}
	return;
}



int int10_emu ( void )
{
  switch (AH) {
		/* set video mode */
    case 0x00 : scrxsize=80;
								scrysize=25;
								curattr=7;
								curx=0;
								cury=0;
								CLEAR_SCREEN;
								return 1;
    /* set cursor shape */
    case 0x01 : return 1;
    /* set cursor position */
    case 0x02 : if (DL<scrxsize&&DH<scrysize) {
									curx=DL;
                  cury=DH;
								}
								return 1;
    /* read cursor position */
    case 0x03 : AX=0;
                CX=0;
                DH=cury;
                DL=curx;
								return 1;
    /* scroll up window */
    case 0x06 : scrollup_window(CL,CH,DL,DH,BH,AL);
                return 1;
    /* scroll down window */
    case 0x07 : scrolldown_window(CL,CH,DL,DH,BH,AL);
                return 1;
    /* read character and attribute at cursor position */
    case 0x08 : AX=*((word*)0xB8000+curx+scrxsize*cury);
							return 1;
		/* print ASCII char/attr */
		case 0x09 : {
									word *p=(word*)0xB8000+curx+scrxsize*cury;
									int n=CX;
									while (n--) *(p++)=AL+(BL<<8);
									return 1;
								}
		/* print ASCII char */
		case 0x0A : {
									byte *p=(byte*)0xB8000+((cury+scrxsize*cury)<<1);
									int n=CX;
									while (n--) {
										*p=AL;
										p+=2;
									}
									return 1;
								}
    /* print character */
    case 0x0E : screen_raw_char(AL,curattr);
								return 1;
    /* read video mode */
    case 0x0F : AL=3;  /* we put 80*25 color mode as respond */
								AH=80;
								BH=0;
								return 1;
    /* get EGA/VGA information */
    case 0x12 : switch (BL) {
									case 0x10 : BX=0x0100;
                  return 1;
                }
#ifdef CONFIG_DEBUG_UNIMPLEMENTED_FUNC
								if (stddebug) fprintf(stddebug,"INT10: function 12h: unknown subfunction BL=%02X\n",BL);
#endif
								return 1;
		/* write string */
		case 0x13 : int10_13_emu();
								return 1;
  }
#ifdef CONFIG_DEBUG_UNIMPLEMENTED_FUNC
  if (stddebug) fprintf(stddebug,"INT10: unimplemented function (CS:IP=%04X:%04X) AX=%04Xh.\n",oldcs,oldip,AX);
#endif
  return 1;
}

#endif

